#include "matrix_lib.h"

/* used to solve a system of equations in the form [A]x[X]=[B] where:
   A is a 2d matrix of size [row , col] 
   B is a 1d matrix of size [row]
   X (sol_X) is the solution 1d matrix of size [row]
   A.col must equal B.row to allow the solution to be computed */

struct matrix_column solution(FILE      *output,
							  struct matrix matrix_A, 
							  struct matrix_column matrix_B) {

	/* decomposition matrices 1 and 2 are used during the solution process */
	/* MAXMAT+1 wide to hold left side (A) and right side (B) of equation all in one array */ 
	struct matrix_decomp { double M[MAXMAT][MAXMAT+1]; } decomp1,decomp2;

	/* sol_X is the solution and will be returned */
	struct matrix_column sol_X;

	int			i,j,
				current_sol_X_row,
				current_row,
				current_col,
				good_swap,
				all_zeros;

	sol_X.rows=0;
	if (matrix_A.cols != matrix_B.rows) {   /* if these are not equal, can not solve for sol_X */
		return(sol_X);
	}

	/* if they are equal, set up everything else */
	sol_X.rows=matrix_A.cols;	

	fprintf(output,"Original system of equations:\n");
	for (i=0;i<matrix_A.rows;i++) {
		fprintf(output,"row %d : ",i);
		for (j=0;j<matrix_A.cols;j++) {
			fprintf(output,"%11.7f X(%d)",matrix_A.M[i][j],j); if (j!=(matrix_A.cols-1)) fprintf(output," + ");
		}
		fprintf(output," = %11.7f\n",matrix_B.M[i]);
	}


	/* go through the solution matrix and get each value */
	for (current_sol_X_row=0;current_sol_X_row<sol_X.rows;current_sol_X_row++) {

		fprintf(output,"\n-----------------------------------------------------------------\n");
		fprintf(output,"Decomposing system of equations to get X(%d) ...",current_sol_X_row);
		
		/* check for all 0's in any one column of matrix_A -> no solution in corresponding sol_X row is possible */ 
		all_zeros=TRUE;
		for (i=0;i<matrix_A.rows;i++) { if (matrix_A.M[i][current_sol_X_row] != 0.0) all_zeros=FALSE; }

		if (all_zeros) {
			fprintf(output,"\nAll 0's in matrix A column %d; no solution is possible for X(%d).\n",current_sol_X_row,current_sol_X_row);
			sol_X.M[current_sol_X_row]=0.0;

		} else { /* do rest of solution */

			/* copy original into working array (decomp2) */
			for (i=0;i<MAXMAT;i++) {
				for (j=0;j<MAXMAT;j++) decomp2.M[i][j]=matrix_A.M[i][j];
				decomp2.M[i][MAXMAT]=matrix_B.M[i];
			}
	
			current_row = 0;  /* used to determine where we are in the decomp arrays; always starts with the first rows */
		
			/* go through the matrix one column at a time and eliminate each coefficient */
			for (current_col=0;current_col<matrix_A.cols;current_col++) {
				if (current_col!=current_sol_X_row) {
				
					/* if current_col does not equal the current_sol_X_row (so we don't divide out the value we want) then 
					      eliminate that coefficient in decomp2 array and place new values in decomp1 */
					fprintf(output,"\nDivide each row by coefficient %d (if not equal to 0)...\n",current_col);
					for (i=current_row;i<matrix_A.rows;i++) {

						/* if the the coefficient in the current_col != 0, divide it */ 
						if (decomp2.M[i][current_col]!=0.0) {
						
							for (j=0;j<=MAXMAT;j++) decomp1.M[i][j]=decomp2.M[i][j]/decomp2.M[i][current_col];
							fprintf(output,"r%d = r%d / r%d(%d) :",i,i,i,current_col);

						} else {  /* it is 0; can not divide - move everything into decomp1 */

							for (j=0;j<=MAXMAT;j++) decomp1.M[i][j]=decomp2.M[i][j];
							fprintf(output,"r%d = r%d (no div):",i,i);
						}

						/* print out current matrix makeup */
						for (j=0;j<matrix_A.cols;j++) {
							fprintf(output,"%11.7f X(%d)",decomp1.M[i][j],j); 
							if (j!=(matrix_A.cols-1)) fprintf(output," + ");
						}
						fprintf(output," = %11.7f\n",decomp1.M[i][MAXMAT]);

					}

					/* move all non-zero rows to top */
					/* partial pivoting here to allow sparse arrays (0 coefficients) to be solved */
					/* if a row already has a zero coefficient, it doesn't need to be subtracted out */
					if (decomp1.M[current_row][current_col]==0.0) {
						fprintf(output,"\nTrying to swap rows ... ");
						good_swap=FALSE;

						for (i=current_row+1;i<matrix_A.rows;i++) {

							if (!good_swap) { /* if not swapped already */
								if (decomp1.M[i][current_col]!=0.0) { /* and this column coeff. is non-zero, swap them */

									/* temp swap value is decomp2.M[0] */
									for (j=0;j<=MAXMAT;j++) { 
										decomp2.M[0][j]=decomp1.M[i][j];
										decomp1.M[i][j]=decomp1.M[current_row][j];
										decomp1.M[current_row][j]=decomp2.M[0][j];
									}

									fprintf(output,"swapped row %d with row %d...\n",current_row,i);
									good_swap=TRUE;
								}
							}
						}

						/* if any swap occurred, print out the new matrix makeup */
						if (good_swap) {
							for (i=current_row;i<matrix_A.rows;i++) {
								fprintf(output,"r%d (new) =      :",i);
								for (j=0;j<matrix_A.cols;j++) {
									fprintf(output,"%11.7f X(%d)",decomp1.M[i][j],j); 
									if (j!=(matrix_A.cols-1)) fprintf(output," + ");
								}
								fprintf(output," = %11.7f\n",decomp1.M[i][MAXMAT]);
							}
						}
					}
			
					/* subtract each row from the current top row to get n-1 rows */
					fprintf(output,"\nTrying to subtract : current row = current row - row %d ...\n",current_row);
					for (i=current_row+1;i<matrix_A.rows;i++) {

						/* if decomp1.M[i][j] != 0, then subtract */
						if (decomp1.M[i][current_col]!=0.0) {

							/* move subtracted results into decomp2 array */
							for (j=0;j<=MAXMAT;j++) decomp2.M[i][j]=decomp1.M[i][j]-decomp1.M[current_row][j];
							fprintf(output,"r%d = r%d - r%d    :",i,i,current_row);
					
						} else {  /* already 0; can not subtract - just copy it into decomp2 */

							for (j=0;j<=MAXMAT;j++) decomp2.M[i][j]=decomp1.M[i][j];
							fprintf(output,"r%d = r%d (no sub):",i,i);
						}

						/* print out current matrix makeup */
						for (j=0;j<matrix_A.cols;j++) {
							fprintf(output,"%11.7f X(%d)",decomp2.M[i][j],j); 
							if (j!=(matrix_A.cols-1)) fprintf(output," + ");
						}
						fprintf(output," = %11.7f\n",decomp2.M[i][MAXMAT]);
					}

					current_row++;
				}
			}

			/* get solution by dividing both sides by the only remaining coefficient */
			sol_X.M[current_sol_X_row]=decomp2.M[matrix_A.rows-1][MAXMAT]/decomp2.M[matrix_A.rows-1][current_sol_X_row];

			fprintf(output,"X(%d) = %11.7f / %11.7f = %11.7f\n",
				current_sol_X_row,decomp2.M[matrix_A.rows-1][MAXMAT],
				decomp2.M[matrix_A.rows-1][current_sol_X_row],
				sol_X.M[current_sol_X_row]);

		} /* end non-zero */

	}
	return(sol_X);

}

/* ------------------------------------------------------------------------------- */
/* ------- transpose will give the transpose of the input matrix ----------------- */
/* ------------------------------------------------------------------------------- */
struct matrix transpose(struct matrix matrix_in)
{
	int i,j;
	struct matrix matrix_out;

	for (i=0;i<matrix_in.rows;i++)
	  for (j=0;j<matrix_in.cols;j++)
		 matrix_out.M[i][j]=matrix_in.M[j][i];
	matrix_out.rows=matrix_in.cols;
	matrix_out.cols=matrix_in.rows;

	return(matrix_out);
}

/* ------------------------------------------------------ */
/* ---- trace returns the trace of a square matrix ------ */
/* ------------------------------------------------------ */
double trace(struct matrix matrix_in)
{
  int i;
  double value=0.0;

  if (matrix_in.rows != matrix_in.cols)
	 return(value);

  for (i=0;i<matrix_in.rows;i++)
	 value = value + matrix_in.M[i][i];

  return(value);
}

/* ------------------------------------------------------------ */
/* ---- mat_mult will mutiply two conformable matrices -------- */
/* ------------------------------------------------------------ */
struct matrix matrix_multiply(struct matrix matrix1,
					struct matrix matrix2)
{
	int i,j,k,l,m,n;
	struct matrix matrix3;

	matrix3.rows=0;
	matrix3.cols=0;

	if ( matrix1.cols != matrix2.rows )
	  return(matrix3);

	l            = matrix1.rows;
	m            = matrix1.cols; /* =matrix2.rows */
	n            = matrix2.cols;
	matrix3.rows = l;
	matrix3.cols = n;

	for (i=0;i<l;i++)
	  for (j=0;j<n;j++)
	  {
		 matrix3.M[i][j]=0.0;
		 for (k=0;k<m;k++)
			matrix3.M[i][j] = matrix3.M[i][j] + (matrix1.M[i][k] * matrix2.M[k][j]);
	  };

	return(matrix3);
}

/* ------------------------------------------------------------------------------------ */
/* ------ cofactor returns the cofactor of element row_pos,col_pos of matrix_in ------- */
/* ------------------------------------------------------------------------------------ */
double cofactor(struct matrix matrix_in,
		int           row_pos,
		int           col_pos)
{
	double value=0.0,power;
	int i;

	if (matrix_in.rows != matrix_in.cols) return(value);
	if (matrix_in.rows < 3)               return(value);

	power=1.0; for (i=0;i<(row_pos+col_pos);i++) power=power*-1.0;
	value=power * minor(matrix_in,row_pos,col_pos);

	return(value);
};


/* ------------------------------------------------------------------- */
/* ------- det will give the determinate of a square matrix ---------- */
/* ------------------------------------------------------------------- */
double det(struct matrix matrix_in)
{
	double value;
	int i,order;

	value=0.0;
	if (matrix_in.rows != matrix_in.cols)
	  return(value);
	order=matrix_in.rows;

	if (order < 2)
	  return(value);

	if (order == 2)
	  value = (matrix_in.M[0][0]*matrix_in.M[1][1]) - (matrix_in.M[0][1]*matrix_in.M[1][0]);

	if (order > 2)
	{
	  value=0.0;
	  for (i=0;i<order;i++)
		 value = value + (matrix_in.M[i][0]*cofactor(matrix_in,i,0));
	};

	return(value);
}

/* ---------------------------------------------------------------------------------------- */
/* -------minor returns the minor of matrix_in from position row_pos,col_pos -------------- */
/* --------------------------------------------------------+------------------------------- */
/* example : matrix_in: 1 2 3  row_pos :3  | 2 3 | = 12-15 |  row_pos :3  | 1 3 | = 6 - 12  */
/*		               	4 5 6  col_pos :1  | 5 6 |  = -3   |  col_pos :2  | 4 6 |  = -6     */
/*		               	7 8 9                              |                                */
/* --------------------------------------------------------+------------------------------- */
double minor(struct matrix matrix_in,
			int           row_pos,
			int           col_pos)
{
  int cur_row,cur_col,row_out,col_out,order_in,order_out;
  double value=0.0;
  struct matrix minor_matrix;

  if (matrix_in.rows != matrix_in.cols) return(value);
  order_in=matrix_in.rows;
  order_out=order_in-1;
  minor_matrix.rows=order_out;
  minor_matrix.cols=order_out;

  row_out=0;
  for (cur_row=0;cur_row<order_in;cur_row++)
  {
	 col_out=0;
	 if (cur_row!=row_pos)
	 {
		for (cur_col=0;cur_col<order_in;cur_col++)
		{
			if (cur_col!=col_pos)
			{
				minor_matrix.M[row_out][col_out]=matrix_in.M[cur_row][cur_col];
				col_out++;
			};
		};
		row_out++;
	 };
  };
  value=det(minor_matrix);

  return(value);
}


/* ------------------------------------------------------------------- */
/* ------- adjoint will give the adjoint of a square matrix ---------- */
/* ------------------------------------------------------------------- */
struct matrix adjoint(struct matrix matrix_in)
{
	struct matrix adjoint_matrix,transposed;
	int i,j,order;

	order=matrix_in.rows;
	transposed.rows=0;
	transposed.cols=0;

	if (matrix_in.rows != matrix_in.cols)
	  return(transposed);

	adjoint_matrix.rows=order;
	adjoint_matrix.cols=order;
	for (i=0;i<order;i++)
	  for (j=0;j<order;j++)
		 adjoint_matrix.M[i][j]=cofactor(matrix_in,i,j);
	transposed=transpose(adjoint_matrix);

	return(transposed);
}

/* ------------------------------------------------------------------- */
/* ------- invert will give the inverse of a square matrix ----------- */
/* ------------------------------------------------------------------- */
struct matrix inverse(struct matrix matrix_in)
{
	 struct matrix adjoint_matrix,inver;
	 double determinate;
	 int i,j,order;

	 order=matrix_in.rows;
	 inver.rows=0;
	 inver.cols=0;

	 if (matrix_in.rows != matrix_in.cols)
		return(inver);
	 determinate=det(matrix_in);
	 if (fabs(determinate)<0.00000001)
		return(inver);

	 adjoint_matrix=adjoint(matrix_in);

	 for (i=0;i<adjoint_matrix.rows;i++)
		for (j=0;j<adjoint_matrix.cols;j++)
		  inver.M[i][j] = adjoint_matrix.M[i][j]/determinate;
	 inver.rows=order;
	 inver.cols=order;

	 return(inver);
}

/* ----------------------------------------------------- */
/* ------- sqr returns the invalue squared ------------- */
/* ----------------------------------------------------- */
double sqr(double invalue)
{
  return(invalue*invalue);
}

/* -------------------------------------------------- */
/* ------- print_matrix will print a matrix --------- */
/* -------------------------------------------------- */
int print_matrix(FILE          *output,
		 struct matrix matrix_in)
{
  int i,j;

  if ((matrix_in.rows == 0) || (matrix_in.cols==0))
	 return(-1);

  fprintf(output,"+");
  for (j=0;j<matrix_in.cols;j++) fprintf(output,"-----------------");
  fprintf(output,"+\n");

  for (i=0;i<matrix_in.rows;i++)
  {
	  fprintf(output,"|");
	  for (j=0;j<matrix_in.cols;j++) fprintf(output,"%15.5f  ",matrix_in.M[i][j]);
	  fprintf(output,"|\n");
  };

  fprintf(output,"+");
  for (j=0;j<matrix_in.cols;j++) fprintf(output,"-----------------");
  fprintf(output,"+\n");

  return(0);
}

int print_matrix_column(FILE *output,
						struct matrix_column matrix_in) {
	int i;
	if (matrix_in.rows == 0) return(-1);
	
	fprintf(output,"+-----------------+\n");
	for (i=0;i<matrix_in.rows;i++)  fprintf(output,"|%15.5f  |\n",matrix_in.M[i]);
	fprintf(output,"+-----------------+\n");

	return(0);
}
	